#! /usr/bin/env python

# bdf.py -- PBD 2015/06

# This class provides access to EVLA/ALMA Binary Data Format (BDF)
# files.  The approach used here is based loosely on 'bdfparse.py' 
# originally by Peter Williams.  So far this is somewhat targeted
# to EVLA/WIDAR data files, it may not handle all the ALMA variants.

from lxml import etree, objectify

import os
import sys
import string
import re
import mmap
import math
import numpy
from copy import deepcopy

from .mime import MIMEPart, MIMEHeader

def basename_noext(path):
    return os.path.basename(os.path.splitext(path)[0])

# TODO find a better way to get the namespace automatically?
_ns = '{http://Alma/XASDM/sdmbin}'
def _stripns(tag):
    return re.sub('{.+}','',tag)

def ant2bl(i,j=None):
    """Returns baseline index for given antenna pair.  Will accept
    two args, or a list/tuple/etc.  Uses 0-based indexing"""
    if j is None:
        (a1,a2) = sorted(i[:2])
    else:
        (a1,a2) = sorted((i,j))
    # could raise error if a2==a1, either are negative, etc
    return (a2*(a2-1))/2 + a1

def bl2ant(i):
    """Returns antenna pair for given baseline index.  All are 0-based."""
    a2 = int(0.5*(1.0+math.sqrt(1.0+8.0*i)))
    a1 = i - a2*(a2-1)/2
    return a1, a2

class BDF(object):
    """
    Class representing a single BDF file.  For example:

        b = bdf.BDF('uid____evla_bdf_1433189755525')

    Individual integration data is returned as BDFIntegration objects via
    either b.get_integration(idx) or b[idx].  Other useful methods include:

        b.basebands      # list of baseband ids
        b.spws           # list of spectral windows per baseband
        b.numAntenna     # number of antennas
        b.numBaseline    # number of baselines
        b.numIntegration # number of integrations in file
        b.sdmDataHeader  # lxml objectify version of full header

    """

    def __init__(self, fname):
        self.fname = fname
        self.fp = open(fname, 'r')
        self.mmdata = mmap.mmap(self.fp.fileno(), 0, mmap.MAP_PRIVATE,
                mmap.PROT_READ)
        self.read_mime()
        self.parse_spws()

    # Size in bytes of each data element. In principle, the crossData 
    # type needs to be read from the headers while all others have 
    # pre-set values...
    bin_dtype_size = {
            'flags':           4, # INT32
            'actualTimes':     8, # INT64
            'actualDurations': 8, # INT64
            'zeroLags':        4, # FLOAT32
            'autoData':        4, # FLOAT32
            'crossData':       4, # FLOAT32 but should be determined from file
            }

    # Basic data type for each array.  Note, does not necessarily
    # match up with the sizes above due to how BDF defines the 'size'
    # attribute..
    bin_dtype = {
            'autoData':        numpy.float32,
            'crossData':       numpy.complex64,
            }

    def read_mime(self,full_read=False):
        self.fp.seek(0,0) # Go back to start
        if not self.fp.readline().startswith('MIME-Version:'):
            raise RuntimeError('Invalid BDF: missing MIME-Version')

        # First we need to read and parse only the main XML header in order
        # to get sizes of the binary parts.  Note, the info stored in 
        # self.bin_size is in bytes, rather than the weird BDF units.
        mime_hdr = MIMEPart(self.fp).hdr
        self.top_mime_bound = mime_hdr.boundary
        sdmDataMime = MIMEPart(self.fp,boundary=self.top_mime_bound)
        if sdmDataMime.loc != 'sdmDataHeader.xml':
            raise RuntimeError('Invalid BDF: missing sdmDataHeader.xml')
        self.sdmDataHeader = objectify.fromstring(sdmDataMime.body)
        self.bin_size = {}
        self.bin_axes = {}
        for e in self.sdmDataHeader.iter():
            if 'size' in e.attrib.keys() and 'axes' in e.attrib.keys():
                binname = _stripns(e.tag)
                self.bin_size[binname] = int(e.attrib['size']) \
                        * self.bin_dtype_size[binname]
                self.bin_axes[binname] = e.attrib['axes'].split()

        # For EVLA, we can read the first integration, note the file offset
        # in order to determine the size, then seek to each integration as
        # requested, rather than parsing the whole file here.
        if 'EVLA' in mime_hdr['Content-Description'][0] and not full_read:
            self.offset_ints = self.fp.tell() # Offset in file to first integ
            self.mime_ints = [MIMEPart(self.fp,
                boundary=self.top_mime_bound,
                binary_size=self.bin_size,
                recurse=True),]
            # Compute size of each integration section:
            self.size_ints = self.fp.tell() - self.offset_ints
            numints = int((os.path.getsize(self.fname)-self.offset_ints)/self.size_ints)
            self.mime_ints += [None,]*(numints-1)

        # This is the more general way to do it that does not assume
        # each integration (including XML and MIME headers) has the 
        # same size in the file.  In this case, go back to the beginning 
        # and parse the whole MIME structure to map it out.
        else:
            self.fp.seek(0,0) # reset to start again
            full_mime = MIMEPart(self.fp,
                    recurse=True, binary_size=self.bin_size)
            self.mime_ints = full_mime.body[1:]

    def _raw(self,idx):
        if self.mime_ints[idx] is not None: return self.mime_ints[idx]
        # Need to read this one
        self.fp.seek(self.offset_ints + idx*self.size_ints, 0)
        #self.mime_ints[idx] = self.read_mime_part(boundary=self.top_mime_bound,recurse=True)
        self.mime_ints[idx] = MIMEPart(self.fp, 
                boundary=self.top_mime_bound, 
                binary_size=self.bin_size,
                recurse=True)
        return self.mime_ints[idx]

    @property
    def projectPath(self):
        return self.sdmDataHeader.attrib['projectPath']

    @property
    def numIntegration(self):
        return len(self.mime_ints)

    @property
    def numAntenna(self):
        return int(self.sdmDataHeader.numAntenna)

    @property
    def numBaseline(self):
        return (self.numAntenna*(self.numAntenna-1))/2

    @property
    def startTime(self):
        return float(self.sdmDataHeader.startTime)/86400.0e9

    def parse_spws(self):
        self.basebands = []
        self.spws = []
        # Offsets into the cross and data arrays where this spw
        # can be found.  This info could be reconstructed later
        # but seems convenient to do it here.
        cross_offset = 0
        auto_offset = 0
        self.spws = []
        for bb in self.sdmDataHeader.dataStruct.baseband:
            bbname = bb.attrib['name']
            self.basebands.append(bbname)
            for spw_elem in bb.spectralWindow:
                # Build a list of spectral windows for each baseband
                spw = BDFSpectralWindow(spw_elem,cross_offset,auto_offset)
                cross_offset += spw.dsize('cross')
                auto_offset += spw.dsize('auto')
                self.spws.append(spw)

    def get_integration(self,idx):
        return BDFIntegration(self,idx)

    def __getitem__(self,idx):
        return self.get_integration(idx)

    def get_data(self,spwidx='all',type='cross',scrunch=False,
            fscrunch=False,frange=None,trange=None):
        """Returns an array containing all integrations for the specified
        baseband, spw and data type.  If scrunch=True, all integrations
        will be averaged."""
        chidx = -2 # index of spectral channels
        # Figure out ranges:
        if trange is None:
            i0 = 0
            i1 = self.numIntegration
        else:
            i0 = trange[0]
            i1 = trange[1]
        # Read first integration to get shapes, etc
        nsubout = i1 - i0
        subdat = self.get_integration(i0).get_data(spwidx,type)
        if scrunch:
            dshape = subdat.shape
        else:
            dshape = (nsubout,) + subdat.shape
        if fscrunch:
            dshape = dshape[:chidx] + dshape[chidx+1:]
        result = numpy.zeros(dshape,dtype=subdat.dtype)
        for i in range(i0,i1):
            if fscrunch:
                if frange is None:
                    dat = self.get_integration(i).get_data(spwidx,type).mean(chidx)
                else:
                    dat = self.get_integration(i).get_data(spwidx,type).take(range(*frange),axis=chidx).mean(chidx)
            else:
                dat = self.get_integration(i).get_data(spwidx,type)
            if scrunch:
                result += dat
            else:
                result[i-i0] = dat
        if scrunch: 
            result /= float(nsubout)
        return result

class BDFSpectralWindow(object):
    """Class that represents spectral window information present in BDF files,
    including storing appropriate offsets into the main data array.  Should be 
    initialized from the spectralWindow XML element from the main BDF 
    header."""

    def __init__(self, spw_elem, cross_offset=None, auto_offset=None):
        self._attrib = spw_elem.attrib
        self.cross_offset = cross_offset
        self.auto_offset = auto_offset

    @property
    def numBin(self):
        return int(self._attrib['numBin'])

    @property
    def numSpectralPoint(self):
        return int(self._attrib['numSpectralPoint'])

    # These two are not listed in the BDF spec but appear in EVLA data:
    @property
    def sw(self):
        return int(self._attrib['sw'])

    @property
    def swbb(self):
        return self._attrib['swbb']

    # Should uniquely identify a sw?
    @property
    def name(self):
        return self.swbb + '-' + str(self.sw)

    def npol(self,type):
        """Return number of polarization array elements for the given data
        type (cross or auto)."""
        try:
            if type[0].lower()=='c':
                return len(self._attrib['crossPolProducts'].split())
            elif type[0].lower()=='a':
                # Good enough?
                l = len(self._attrib['sdPolProducts'].split())
                return 4 if l==3 else l # 3==4 in BDF math! :)
        except KeyError:
            return 0

    def dshape(self,type):
        """Return shape tuple of data array for this spectral window, 
        in number of data elements (real for auto, complex for cross)."""
        return (self.numBin, self.numSpectralPoint, self.npol(type))

    def dsize(self,type):
        """Return size of data array for this spectral window, in number of
        data elements (real for auto, complex for cross)."""
        return numpy.product(self.dshape(type))

    @staticmethod
    def dims_match(spwlist,type):
        """Given a list of BDFSpectralWindow objects, return true if all
        of them have consistent array dimensions."""
        if len(spwlist)==1: 
            return True
        for spw in spwlist[1:]:
            if spwlist[0].dshape(type) != spw.dshape(type):
                return False
        return True

class BDFIntegration(object):
    """
    Describes and holds data for a single intgration within a BDF file.
    This should be derived from an existing BDF object using 
    get_integration() or indexing, ie:

        b = bdf.BDF('some_file')

        # Get the 5th integration, these two are equivalent:
        i = b.get_integration(5)
        i = b[5]

        # Read the cross-corr data array for spectral window 0 
        dat = i.get_data(0)

    Other potentially useful info:

        i.basebands            # list of baseband IDs
        i.spws                 # dict of spws per baseband
        i.numAntenna           # obvious
        i.numBaseline          # "
        i.sdmDataSubsetHeader  # lxml objectify version of full sub-header

    """

    def __init__(self,bdf,idx):
        # Get the main header
        self.sdmDataSubsetHeader = objectify.fromstring(
                bdf._raw(idx).body[0].body)
        # Copy some info from the BDF headers
        self.basebands = bdf.basebands
        self.spws = bdf.spws
        self.bin_axes = bdf.bin_axes
        self.numAntenna = bdf.numAntenna
        self.numBaseline = bdf.numBaseline
        # Get the binary data
        self.data = {}
        for m in bdf._raw(idx).body[1:]:
            btype = basename_noext(m.loc)
            bsize = bdf.bin_size[btype] # size of the binary blob in bytes
            baxes = self.bin_axes[btype]
            # Determine outer step size for the array, either baselines
            # antennas of baseline+antenna.  We can't apply the other
            # dimensions here because the number of elements can vary
            # per spw.
            if baxes[0]=='BAL' and baxes[1]=='ANT':
                shape=(self.numBaseline+self.numAntenna,-1)
            elif baxes[0]=='BAL':
                shape=(self.numBaseline,-1)
            elif baxes[0]=='ANT':
                shape=(self.numAntenna,-1)
            else:
                shape=(-1,) # Don't know what to do, just leave flat array
            self.data[btype] = numpy.frombuffer(bdf.mmdata[m.body:m.body+bsize],
                    dtype=bdf.bin_dtype[btype]).reshape(shape)

    @property
    def projectPath(self):
        return self.sdmDataSubsetHeader.attrib['projectPath']

    @property
    def time(self):
        return float(self.sdmDataSubsetHeader.schedulePeriodTime.time)/86400.0e9

    @property
    def interval(self):
        return float(self.sdmDataSubsetHeader.schedulePeriodTime.interval)*1e-9

    def get_data(self,spwidx='all',type='cross'):
        """
        Return the data array for the given subset.  Inputs are:

            spwidx:    spw index within file
            type:      'cross' or 'auto' (default 'cross')

        The returned array shape is (nBl/nAnt, nBin, nSpp, nPol).

        If spwidx is 'all' and the array dimensions of all spectral
        windows match, all will be returned in a single array.  In this case
        the returned dimensions will be (nBl/nAnt, nSpw, nBin, nSpp, nPol).
        """
        if type[0].lower()=='c': 
            loc = 'crossData'
        elif type[0].lower()=='a': 
            loc = 'autoData'
        else:
            raise RuntimeError('Unsupported data type')
        if spwidx=='all':
            if not BDFSpectralWindow.dims_match(self.spws,type):
                raise RuntimeError('BDFIntegration: ' +
                        'mixed array dimensions, spws must be ' +
                        'retrieved indivdually')
            dshape = (-1, len(self.spws)) + self.spws[0].dshape(type)
            return self.data[loc].reshape(dshape)
        spw = self.spws[spwidx]
        if loc == 'crossData': 
            offs = spw.cross_offset
        elif loc == 'autoData': 
            offs = spw.auto_offset
        else:
            raise RuntimeError('Unsupported data type')
        dsize = spw.dsize(type)
        dshape = (-1,) + spw.dshape(type)
        return self.data[loc][:,offs:offs+dsize].reshape(dshape)

class BDFWriter(object):
    """
    Write a BDF file.
    """
    def __init__(self, fname, bdf=None):
        """Init BDFWrite with output filename (fname).  If the bdf
        argument contains a BDF object, its header is copied for the
        output file.  Otherwise the BDFWrite.sdmDataHeader needs to 
        be populated."""
        self.fname = fname
        self.fp = None
        self.curidx = 1
        self.mb1 = "MIME_boundary-1"
        self.mb2 = "MIME_boundary-2"
        self.len0 = 0
        self.len1 = 0
        self.len2 = 0
        self.sdmDataHeader = None
        if bdf is not None: 
            self.sdmDataHeader = deepcopy(bdf.sdmDataHeader)

    def write_header(self):
        """Open output and write the current header contents."""
        self.fp = open(self.fname,'w')
        tophdr = MIMEHeader()
        tophdr['MIME-Version'] = ['1.0',]
        tophdr['Content-Type'] = ['multipart/mixed', 'boundary='+self.mb1]
        tophdr['Content-Description'] = [
                'EVLA/CORRELATOR/WIDAR/FULL_RESOLUTION',]
        # How do we generate a new unique name?
        nsxl = self.sdmDataHeader.nsmap['xl']
        uid = self.sdmDataHeader.dataOID.attrib['{%s}href'%nsxl][5:]
        tophdr['Content-Location'] = ['http://evla.nrao.edu/wcbe/XSDM' + uid,]
        self.fp.write(tophdr.tostring() + '\n')

        self.fp.write('--' + self.mb1 + '\n')
        xhdr = MIMEHeader()
        xhdr['Content-Type'] = ['text/xml', 'charset=utf-8']
        xhdr['Content-Location'] = ['sdmDataHeader.xml',]
        self.fp.write(xhdr.tostring() + '\n')
        self.fp.write(etree.tostring(self.sdmDataHeader,
            standalone=True,encoding='utf-8') + '\n')

    def write_integration(self,bdf_int):
        """Input is a BDFIntegration object.  The projectPath will be updated
        so that it is consistent for the file being written but otherwise
        no changes are made to the contents."""
        tophdr = MIMEHeader()
        tophdr['Content-Type'] = ['multipart/related', 'boundary='+self.mb2]
        tophdr['Content-Description'] = ['data and metadata subset',]

        ppidx = self.sdmDataHeader.attrib['projectPath'] + '%d/' % self.curidx

        hdr = MIMEHeader()
        hdr['Content-Type'] = ['text/xml', 'charset=utf-8']
        hdr['Content-Location'] = [ppidx + 'desc.xml']
        if self.len0==0:
            self.len0 = len(hdr.tostring()) + 12
        nxpad = self.len0 - len(hdr.tostring())
        if nxpad<0: 
            raise RuntimeError('nxpad(0)<0')
        hdr['X-pad'] = ['*'*nxpad,]

        # Update subhdr with new path
        subhdr = deepcopy(bdf_int.sdmDataSubsetHeader)
        subhdr.attrib['projectPath'] = ppidx
        nsxl = subhdr.nsmap['xl']
        dtypes = []
        mhdr = {}
        for dtype in ('crossData', 'autoData'):
            try:
                loc = ppidx + dtype + '.bin'
                getattr(subhdr,dtype).attrib['{%s}href'%nsxl] = loc
                dtypes += [dtype,]
                mhdr[dtype] = MIMEHeader()
                mhdr[dtype]['Content-Type'] = ['application/octet-stream']
                mhdr[dtype]['Content-Location'] = [loc]
            except AttributeError:
                pass

        # Figure out how much X-pad to add
        subhdr_str = etree.tostring(subhdr, standalone=True, encoding='utf-8')
        if self.len1==0:
            self.len1 = len(subhdr_str) + len(mhdr[dtypes[0]].tostring()) + 50
        nxpad = self.len1 - (len(subhdr_str) + len(mhdr[dtypes[0]].tostring()))
        mhdr[dtypes[0]]['X-pad'] = ['*'*nxpad,]

        # Assumes at most 2 data types.. TODO make more general
        if len(dtypes)>1:
            if self.len2==0:
                self.len2 = len(mhdr[dtypes[1]].tostring()) + 12
            nxpad = self.len2 - len(mhdr[dtypes[1]].tostring())
            mhdr[dtypes[1]]['X-pad'] = ['*'*nxpad,]

        # TODO could check that data sizes match up with header info..

        # Now write it all out..
        self.fp.write('--' + self.mb1 + '\n')
        self.fp.write(tophdr.tostring()+'\n')

        # XML subheader part
        self.fp.write('--' + self.mb2 + '\n')
        self.fp.write(hdr.tostring() + '\n')
        self.fp.write(subhdr_str)

        # Data parts
        for dtype in dtypes:
            self.fp.write('\n--' + self.mb2 + '\n')
            self.fp.write(mhdr[dtype].tostring() + '\n')
            self.fp.write(bdf_int.data[dtype])

        # Close out mime
        self.fp.write('\n--' + self.mb2 + '--\n')
        self.curidx += 1

    def close(self):
        self.fp.write('--' + self.mb1 + '--\n')
        self.fp.close()




# scan.py
#
# higher-level class to gather per-scan info from SDM and BDF sets

import string
import numpy
import os.path

from .bdf import BDF, ant2bl, bl2ant

def uid2fname(s):
    """Convert uid URL to file name (mainly for BDFs)."""
    return s.translate(string.maketrans(':/','__'))

def sdmarray(s):
    """Convert an array-valued SDM entry (string) into a numpy array."""
    fields = str(s).split()
    ndim = int(fields[0])
    dims = tuple(map(int,fields[1:ndim+1]))
    return numpy.array(fields[ndim+1:]).reshape(dims)

class Scan(object):
    """
    Represents a single scan as part of a SDM/BDF dataset.  Convenience
    interface to open the BDF, get useful metadata, etc.
    """
    def __init__(self, sdm, scanidx):
        """
        sdm is the SDM object.
        scanidx is the index into the Main table.
        """
        self.sdm = sdm
        self.idx = scanidx
        self._bdf = None
        bdfdir = sdm.bdfdir if sdm.bdfdir else os.path.join(sdm.path, 'ASDMBinary')
        try:
            self._bdf_fname = os.path.join(bdfdir,
                                           uid2fname(sdm['Main'][self.idx].dataUID.EntityRef.get('entityId')))
        except AttributeError:
            self._bdf_fname = os.path.join(bdfdir,
                                           uid2fname(sdm['Main'][self.idx].dataOid.EntityRef.get('entityId')))

    @property
    def bdf(self):
        if self._bdf is None:
            self._bdf = BDF(self._bdf_fname)
        return self._bdf

    @property
    def _main(self):
        """Convenience interface to the SDM Main table row."""
        return self.sdm['Main'][self.idx]

    @property
    def _scan(self):
        """Convenience interface to the SDM Scan table row."""
        return self.sdm['Scan'][self.idx]

    @property
    def _subscan(self):
        """Convenience interface to the SDM Subscan table row."""
        return self.sdm['Subscan'][self.idx]

    @property
    def _config(self):
        """Convenience interfact to the SDM ConfigDescription row."""
        return self.sdm['ConfigDescription'][self._main.configDescriptionId]

    @property
    def source(self):
        return self._scan.sourceName

    @property
    def intents(self):
        """Return the list of intents for this scan."""
        return list(sdmarray(self._scan.scanIntent))

    @property
    def antennas(self):
        """Return the list of antenna names for this scan."""
        sdm_ants = sdmarray(self._config.antennaId)
        return [self.sdm['Antenna'][a].name for a in sdm_ants]

    @property
    def stations(self):
        """Return the list of station names for this scan."""
        sdm_ants = sdmarray(self._config.antennaId)
        sdm_stns = [self.sdm['Antenna'][a].stationId for a in sdm_ants]
        return [self.sdm['Station'][s].name for s in sdm_stns]

    @property
    def baselines(self):
        """Return the list of antenna pairs for this scan, in BDF ordering."""
        ants = self.antennas
        nant = len(ants)
        nbl = nant*(nant-1)/2
        #return ['%s-%s' % (ants[x[0]], ants[x[1]]) 
        #        for x in map(bl2ant, range(nbl))]
        return [(ants[x[0]], ants[x[1]]) for x in map(bl2ant, range(nbl))]

    @property
    def spws(self):
        """ Return the list of spw names """
        
        return [self.sdm['DataDescription'][dd_id].spectralWindowId for dd_id in sdmarray(self._config.dataDescriptionId)]

    @property
    def reffreqs(self):
        """ List of reference frequencies. One per spw in spws list. """

        return [self.spw(spwn).refFreq for spwn in range(len(self.spws))]

    @property
    def numchans(self):
        """ List of number of channels per spw. One per spw in spws list. """

        return [self.spw(spwn).numChan for spwn in range(len(self.spws))]

    @property
    def chanwidths(self):
        """ List of channel widths. One per spw in spws list. """

        return [self.spw(spwn).chanWidth for spwn in range(len(self.spws))]

    def spw(self,idx):
        """Return the SpectralWindow entry for the given index in this scan."""
        dd_id = sdmarray(self._config.dataDescriptionId)[idx]
        spw_id = self.sdm['DataDescription'][dd_id].spectralWindowId
        return self.sdm['SpectralWindow'][spw_id]





#! /usr/bin/env python

# sdm.py -- P. Demorest, 2015/03

import os.path
from lxml import etree, objectify

from .scan import Scan

_install_dir = os.path.abspath(os.path.dirname(__file__))
_xsd_dir = os.path.join(_install_dir, 'xsd')
_sdm_xsd = os.path.join(_xsd_dir, 'sdm_all.xsd')
# Might want to make the schema location settable somehow?
_sdm_parser = objectify.makeparser(schema=etree.XMLSchema(file=_sdm_xsd))

class SDM(object):
    """
    Top-level class to represent an SDM.

    Init arguments:
      path = path to SDM directory
      bdfdir = different directory to search for bdfs (optional, for pre-archive SDMs)

    Attributes:
      tables = list of tables
      path   = full path to SDM directory

    SDM['TableName'] returns the relevant SDMTable object.
    """
    def __init__(self,path='.',use_xsd=True, bdfdir=''):
        if use_xsd: parser = _sdm_parser
        else: parser = None
        self._tables = {}
        self.path = os.path.abspath(path)
        self.bdfdir = bdfdir
        self._asdmtree = objectify.parse(path+'/ASDM.xml',parser)
        self.asdm = self._asdmtree.getroot()
        for tab in self.asdm.Table:
            self._tables[tab.Name] = sdmtable(tab.Name,path,use_xsd=use_xsd)

    @property
    def tables(self):
        """Return the list of table names"""
        return self._tables.keys()

    def __getitem__(self,key):
        return self._tables[key]

    def scan(self,idx):
        """Return a Scan object for the given scan number."""
        return Scan(self,str(idx))

    def scans(self):
        """Iterate over scans."""
        # List of SDM scan numbers:
        scanidx = [s.scanNumber for s in self['Scan']]
        for idx in scanidx:
            yield self.scan(idx)

    def _update_ASDM(self):
        """Updates the ASDM table with the current number of rows."""
        # TODO could check UIDs as well
        for tab in self.asdm.Table:
            try:
                nrow = len(self[tab.Name])
                tab.NumberRows = nrow
            except TypeError:
                # This is a workaround since len is not yet implemented for
                # binary tables.
                pass

    def write(self,newpath):
        """Write the SDM out to the new path location.  Currently does not
        copy the BDFs or anything else under ASDMBinary."""
        self._update_ASDM()
        if not os.path.exists(newpath): os.mkdir(newpath)
        # Write ASDM.xml
        objectify.deannotate(self._asdmtree,cleanup_namespaces=True)
        self._asdmtree.write(newpath+'/ASDM.xml',
                encoding='utf-8', pretty_print=True, standalone=True)
        # Call each table's write method for the rest
        for tab in self.tables: self[tab].write(newpath)

def sdmtable(name,path,*args,**kwargs):
    """
    Return the correct type of SDM table object (binary or XML).
    """
    fnamebase = os.path.join(path,str(name))
    if os.path.exists(fnamebase + '.xml'):
        return SDMTable(name,path,*args,**kwargs)
    elif os.path.exists(fnamebase + '.bin'):
        return SDMBinaryTable(name,path,*args,**kwargs)
    return None

def decap(s):
    return s[:1].lower() + s[1:] if s else ''

class SDMTable(object):
    """
    Class for an individual SDM table.

    Generally this should not be used directly, but as part of a full
    SDM via the SDM class.  

    Init arguments:
      name = Name of table (not including .xml extension)
      path = Path to SDM directory
      idtag = Name of ID tag in table (defaults to nameId)

    SDMTable[i] returns the i-th row as an lxml objectify object
    """

    # Any non-standard Id tag names can be listed here.
    # Note that for some tables a unique key is not a single tag
    # but is defined as a combination of several tags.  We are
    # just going to ignore this for now and do something that is
    # convenient and seems to work for most EVLA data sets.
    _idtags = {
            'Main': 'scanNumber',
            'Scan': 'scanNumber',
            'Subscan': 'scanNumber',
            }

    def __init__(self,name,path='.',idtag=None,use_xsd=True):
        self.name = name
        if idtag is None:
            try:
                self.idtag = self._idtags[name]
            except KeyError:
                self.idtag = decap(str(name)) + 'Id'
        else:
            self.idtag = idtag
        if use_xsd: parser = _sdm_parser
        else: parser = None
        self._tree = objectify.parse(path+'/'+name+'.xml',parser)
        self._table = self._tree.getroot()

    @property
    def entityId(self):
        """Shortcut to entityId"""
        return self._table.Entity.get('entityId')

    @property
    def containerId(self):
        """Shortcut to ContainerEntity entityId"""
        return self._table.ContainerEntity.get('entityId')

    def __getitem__(self,key):
        if type(key)==int:
            return self._table.row[key]
        else:
            # Search through the table and find the first with
            # the matching id tag:
            for r in self._table.row:
                try:
                    if str(getattr(r,self.idtag)) == key:
                        return r
                except AttributeError:
                    pass
            # No matching rows:
            raise KeyError(key)

    def __len__(self):
        try:
            return len(self._table.row)
        except AttributeError:
            return 0

    def write(self,newpath,fname=None):
        """
        Write the updated XML file to the specified path.  Will be named 
        TableName.xml unless overridden via the fname argument.
        """
        objectify.deannotate(self._tree, cleanup_namespaces=True)
        if fname is None:
            outf = os.path.join(newpath,self.name+'.xml')
        else:
            outf = os.path.join(newpath,fname)
        self._tree.write(outf, encoding='utf-8', pretty_print=True,
                standalone=True)

#class SDMTableRow(objectify.ObjectifiedElement):
#    """
#    In case we want to add any extra functionality to table rows, 
#    could try to get this working.  Will require some lxml-fu, and is
#    not currently implemented.
#    """
#
#    def __str__(self):
#        return str(self.__dict__)
#
#    @property
#    def keys(self):
#        return self.__dict__.keys()

class SDMBinaryTable(object):
    """
    Represents an SDM binary table.  Not really implemented yet, but will
    read the data and write it back out when asked to do so by the main
    SDM class.
    """
    def __init__(self,name,path,use_xsd=None):
        self.name = name
        self._data = open(path+'/'+name+'.bin','r').read()

    def write(self,newpath,fname=None):
        if fname is None:
            outf = os.path.join(newpath,self.name+'.bin')
        else:
            outf = os.path.join(newpath,fname)
        open(outf,'w').write(self._data)





#!/users/claw/miniconda/envs/development/bin/python

# Average SDM/BDF data in time
# PBD 05/2016

import os, sys
import numpy as np
np.errstate(divide='ignore')
import sdmpy
import progressbar

sdmname = sys.argv[1]

sdm = sdmpy.SDM(sdmname)

tavg = 3.0 # in seconds

sdmout = sdmname + '.avg'
bdfoutpath = sdmout + '/ASDMBinary'

os.mkdir(sdmout)
os.mkdir(bdfoutpath)

for scan in sdm.scans():
    print "Processing '%s' scan %s:" % (sdmname, scan.idx)
    try:
        bdf = scan.bdf
    except IOError:
        print "Error reading bdf for scan %s, skipping" % (scan.idx,)
        continue
    # TODO maybe rename BDFs...
    # Also this assumes averaging time is the same during the whole
    # BDF.  This is always true for VLA data.
    bdfoutname = bdfoutpath + '/' + os.path.basename(scan._bdf_fname)
    navg = int(tavg / bdf[0].interval)
    nout = int(bdf.numIntegration / navg)
    delta_t = int((navg/2.0)*bdf[0].interval*1e9) # ns
    # Set up for output BDF, copying header info from the input BDF
    bdfout = sdmpy.bdf.BDFWriter(bdfoutname, bdf=bdf)
    bdfout.write_header()
    bar = progressbar.ProgressBar()
    for i in bar(range(nout)):
        bdfint = bdf[i*navg]
        count = {}
        for dtype in bdfint.data.keys():
            bdfint.data[dtype] = bdfint.data[dtype].copy()
            count[dtype] = np.zeros(bdfint.data[dtype].shape)
            count[dtype] += bdfint.data[dtype] != 0.0
        for j in range(1,navg):
            for dtype in bdfint.data.keys():
                dat = bdf[i*navg+j].data[dtype]
                bdfint.data[dtype] += dat
                count[dtype] += dat != 0.0
        for dtype in bdfint.data.keys():
            bdfint.data[dtype] /= count[dtype]
            bdfint.data[dtype][np.where(count[dtype]==0.0)] = 0.0
        # update timestamp and interval
        bdfint.sdmDataSubsetHeader.schedulePeriodTime.time += delta_t
        bdfint.sdmDataSubsetHeader.schedulePeriodTime.interval *= navg
        bdfout.write_integration(bdfint)
    bdfout.close()

    # update SDM entries with corect number of integrations, etc.
    scan._main.numIntegration = nout
    scan._main.dataSize = os.path.getsize(bdfoutname)
    scan._subscan.numIntegration = nout
    scan._subscan.numSubintegration = ('1 %d' % nout) + ' 0'*nout

# Write out new SDM tables
sdm.write(sdmout)




#! /usr/bin/env python

# mime.py -- PBD 2016/04

# This class provdes MIME-parsing functionality suitable for reading
# EVLA/ALMA Binary Data Format (BDF) files, and SDM binary tables.  It
# is not meant to be a totally general MIME reader.  In particular,
# contents of binary type are not returned directed, rather an offset
# into the file is given.

import string
from collections import OrderedDict

class MIMEHeader(OrderedDict):
    # MIMEHeader is a dict with keys equal to the mime header keywords
    # and values equal to lists of entries.  Helper functions for parsing
    # or generating the MIME-format headers are collected here.

    @property
    def boundary(self):
        if self['Content-Type'][0].startswith('multipart/'):
            for v in self['Content-Type']:
                if v.startswith('boundary='):
                    return v[v.index('=')+1:]
        return None

    def addline(self,line):
        """
        Given a single line from a mime header, split it into key/val and
        add it to the dict.
        """
        idx = line.index(':')
        key = line[:idx]
        vals = map(string.strip, line[idx+1:].split(';'))
        self[key] = vals

    @staticmethod
    def _asline(key,val):
        """Convert given key and value list to MIME header line."""
        return key + ': ' + string.join(val, '; ') + '\n'

    def tostring(self,key=None):
        """
        Return contents as in MIME-header format.  If key is given, only
        the line corresponding to the requested key will be returned,
        otherwise the full header will be returned.
        """
        if key is not None:
            return self._asline(key,self[key])
        else:
            out = ''
            for k in self.keys():
                out += self._asline(k,self[k])
            return out

    def __str__(self):
        return self.tostring()

# TODO make a utils.py to hold stuff like this
import os
def basename_noext(path):
    return os.path.basename(os.path.splitext(path)[0])

class MIMEPart(object):
    """
    Class for representing one part of a MIME message.
    Has two member variable:

      hdr  = Dict of MIME header key/value pairs
      body = Body of message.  In our usage, can be a file offset in bytes
             (for binary parts), a string (for text) or a list of MIMEPart
             objects (for multipart).

    The loc property is a shortcut for the Content-Location header
    parameter.

    The type property is a shortcut for Content-Type
    """

    def __init__(self,fp,boundary=None,recurse=False,binary_size=None):
        """
        Read a MIME content part starting at the current file location.
        Return value is a MIMEPart object, which has elements:

            hdr    dict of MIME header key / value pairs

            body   string if Content-Type was 'text/xml', offset into
                   the file if 'application/octet-stream', or list of
                   other MIMEParts for a 'multipart/*'.

        If recurse is True, will read/return the contents of a multipart
        (and any multiparts found at lower levels).  Otherwise will read
        one header/body unit and pointer will be left at the start of 
        the next one (or first sub-part for multiparts).

        binary_size is a dict of sizes of binary components by type.
        For each binary part found, if Content-Location agrees with type,
        the binary data will be skipped over rather than read (this is
        for reading BDF files).  If binary_size is not given, or if an
        unknown type is found, the data must be read to determine its 
        size, however this last part is not implemented yet.
        """
        self.hdr = MIMEHeader({})
        self.body = None
        in_hdr = True
        binary_type = False
        multipart_type = False
        # Note, need to use readline() rather than iterating over file
        # because we need to recover file positions and seek ahead.
        # The "for line in file" type loop reads ahead so is not compatible
        # with this approach.
        while True:

            line = fp.readline()

            # hit EOF
            if line=='':
                return

            # Check for multipart boundary marker
            if boundary is not None:
                if in_hdr:
                    # If we are starting, ignore a 'start' marker,
                    # quit on a 'done' marker
                    if line=='--'+boundary+'\n':
                        continue
                    elif line=='--'+boundary+'--\n':
                        self.hdr = MIMEHeader({})
                        self.body = None
                        return
                else:
                    # This marks the end of a part, rewind so that the 
                    # next part can be parsed, and return results
                    if line.startswith('--' + boundary):
                        fp.seek(-len(line),1)
                        return

            if line=='\n':
                # Got blank line, the next part will be body.  We
                # want to skip it if this is a binary part, otherwise
                # read and return the body.
                in_hdr = False
                if binary_type:
                    # Note the location within the file and skip
                    # ahead by the correct amount.
                    bin_name = basename_noext(self.hdr['Content-Location'][0])
                    self.body = fp.tell()
                    # NOTE if the list of binary sizes is not given, or
                    # if the bin_name is unknown, we should read the data
                    # until the boundary marker is found to determine the 
                    # size.  THIS NEEDS TO BE IMPLEMENTED SOMETIME.  For
                    # now this case will raise an error.
                    if ((binary_size is None) 
                            or (bin_name not in binary_size.keys())):
                        raise RuntimeError("Unknown binary type '%s' found"
                                % bin_name)
                    # Need to add one extra byte for the newline
                    fp.seek(binary_size[bin_name]+1, 1)
                elif multipart_type:
                    if recurse:
                        # Parse the parts and add to a list
                        while True:
                            pmime = MIMEPart(fp, boundary=boundary,
                                        recurse=True,
                                        binary_size=binary_size)
                            if pmime.hdr == {}:
                                return
                            else:
                                self.body.append(pmime)
                continue

            if in_hdr:
                # Still in the header, parse the line as MIME key/val
                self.hdr.addline(line)
                if 'Content-Type' in line:
                    vals = self.hdr['Content-Type']
                    if vals[0].startswith('multipart/'):
                        multipart_type = True
                        boundary = self.hdr.boundary
                        self.body = []
                    elif vals[0] == 'application/octet-stream':
                        binary_type = True
            else:
                if not binary_type:
                    # In body part of a non-binary type
                    if self.body is None: self.body = line
                    else: self.body += line
                else:
                    # Should not really get here, means size calculation
                    # failed or file is otherwise messed up... what to do?
                    raise RuntimeError('MIME parsing failure')

    @property
    def loc(self):
        try:
            return self.hdr['Content-Location'][0]
        except KeyError:
            return None

    @property
    def type(self):
        try:
            return self.hdr['Content-Type'][0]
        except KeyError:
            return None




#!/users/claw/miniconda/envs/development/bin/python

# Split a phase-binned SDM/BDF set into multiple single-bin
# sets.  Currently requires all scans/spws in the set to use the same
# number of bins.
# PBD 06/2016

import os, sys
from collections import namedtuple
import numpy as np
np.errstate(divide='ignore')
import sdmpy
import progressbar

sdmname = sys.argv[1]

sdm = sdmpy.SDM(sdmname)

# Get the number of bins
# TODO clean up spw indexing
bdf0 = sdm.scan(1).bdf
nbin = bdf0.spws[bdf0.basebands[0]][0].numBin

sdmout = []
bdfoutpath = []
for ibin in range(nbin):
    sdmout.append(sdmname + '.bin%04d'%ibin)
    bdfoutpath.append(sdmout[ibin] + '/ASDMBinary')
    os.mkdir(sdmout[ibin])
    os.mkdir(bdfoutpath[ibin])

for scan in sdm.scans():
    print "Processing '%s' scan %s:" % (sdmname, scan.idx)
    try:
        bdf = scan.bdf
    except IOError:
        print "Error reading bdf for scan %s, skipping" % (scan.idx,)
        continue
    bdfoutname = map(lambda x: x+'/'+os.path.basename(scan._bdf_fname), 
            bdfoutpath)
    # Set up for output BDFs, copying header info from the input BDF
    # and changing nbin to 1.
    bdfout = map(lambda x: sdmpy.bdf.BDFWriter(x,bdf=bdf), bdfoutname)
    for ibdf in bdfout: 
        for bb in ibdf.sdmDataHeader.dataStruct.baseband:
            for spw in bb.spectralWindow:
                spw.attrib['numBin'] = '1'
        ibdf.write_header()
    bar = progressbar.ProgressBar()
    for i in bar(range(bdf.numIntegration)):
        fullint = bdf[i]
        dtypes = fullint.data.keys()
        binint = []
        for ibin in range(nbin):
            # This is kind of a kludge to create a new BDFInt-like container
            # to hold the data that will be written out.  May want to clean
            # this up at some point.
            binint.append(namedtuple("NewBDFIntegration",
                ['sdmDataSubsetHeader','data'])(
                    fullint.sdmDataSubsetHeader,
                    {}
                    )
                )
            for dtype in dtypes:
                binint[-1].data[dtype] = None
        for dtype in dtypes:
            for bb in fullint.basebands:
                for ispw in range(len(fullint.spws[bb])):
                    # Axes should be bl/ant, bin, chan, poln
                    data = fullint.get_data(bb,ispw,type=dtype)
                    for ibin in range(nbin):
                        bindata = data.take(ibin,
                                axis=1).reshape((data.shape[0],-1))
                        if binint[ibin].data[dtype] is None:
                            binint[ibin].data[dtype] = bindata.copy()
                        else:
                            binint[ibin].data[dtype] = np.hstack(
                                    (binint[ibin].data[dtype], bindata) 
                                    )
        for ibin in range(nbin):
            bdfout[ibin].write_integration(binint[ibin])
                
    for b in bdfout: b.close()

    # update SDM entries with corect data size (should be same for all bins)
    scan._main.dataSize = os.path.getsize(bdfoutname[0])

# Write out new SDM tables
for s in sdmout: sdm.write(s)




#! /usr/bin/env python

# calib.py -- PBD 2016/05

# Some routines to derive simple calibration solutions directly
# from data arrays.

import numpy as np
from numpy import linalg

from .bdf import ant2bl, bl2ant

def gaincal(data,axis=0,ref=0,avg=[],nit=3):
    """Derives amplitude/phase calibration factors from the data array
    for the given baseline axis.  In the returned array, the baseline
    dimension is converted to antenna.  No other axes are modified.
    Note this internally makes a transposed copy of the data so be 
    careful with memory usage in the case of large data sets.  A list
    of axes to average over before solving can be given in the avg 
    argument (length-1 dimensions are kept so that the solution can be
    applied to the original data)."""
    nbl = data.shape[axis]
    ndim = len(data.shape)
    (check,nant) = bl2ant(nbl)
    if check!=0:
        raise RuntimeError("Specified axis dimension (%d) is not a valid number of baselines" % nbl)
    for a in avg:
        data = data.mean(axis=a,keepdims=True)
    tdata = np.zeros(data.shape[:axis]+data.shape[axis+1:]+(nant,nant),
            dtype=data.dtype)
    for i in range(nbl):
        (a0,a1) = bl2ant(i)
        tdata[...,a0,a1] = data.take(i,axis=axis)
        tdata[...,a1,a0] = np.conj(data.take(i,axis=axis))
    for it in range(nit):
        (wtmp,vtmp) = linalg.eigh(tdata)
        v = vtmp[...,-1].copy()
        w = wtmp[...,-1]
        for i in range(nant):
            tdata[...,i,i] = w*(v.real[...,i]**2 + v.imag[...,i]**2)
    #result = np.sqrt(w[...,-1]).T*v[...,-1].T
    result = np.sqrt(w).T*v.T
    # First axis is now antenna.. refer all phases to reference ant
    result = (result*np.conj(result[ref])/np.abs(result[ref])).T
    # TODO try to reduce number of transposes
    outdims = range(axis) + [-1,] + range(axis,ndim-1)
    return result.transpose(outdims)

def applycal(data,caldata,axis=0,phaseonly=False):
    """Apply the complex gain calibration given in the caldata array
    to the data array.  The baseline/antenna axis must be specified in
    the axis argument.  Dimensions of all other axes must match up 
    (in the numpy broadcast sense) between the two arrays."""
    ndim = len(data.shape)
    nbl = data.shape[axis]
    nant = caldata.shape[axis]
    (check,nant_check) = bl2ant(nbl)
    if check!=0:
        raise RuntimeError("Specified axis dimension (%d) is not a valid number of baselines" % nbl)
    if nant!=nant_check:
        raise RuntimeError("Number of antennas does not match (data=%d, caldata=%d)" % (nant_check, nant))
    if phaseonly:
        caldata = caldata.copy()/abs(caldata)
        caldata[np.where(np.isfinite(caldata)==False)] = 0.0j
    # Modifies data in place.  Would it be better to return a calibrated
    # copy instead of touching the original?
    for ibl in range(nbl):
        # Must be some cleaner way to do this..?
        dslice = (slice(None),)*axis + (ibl,) + (slice(None),)*(ndim-axis-1)
        (a1,a2) = bl2ant(ibl)
        calfac = 1.0 / (caldata.take(a1,axis=axis)
                * caldata.take(a2,axis=axis).conj())
        calfac[np.where(np.isfinite(calfac)==False)] = 0.0j
        data[dslice] *= calfac


